/*
Fix for horizontal stacking weirdness in the RTD theme with Python properties:
https://github.com/readthedocs/sphinx_rtd_theme/issues/1301
*/
.py.property {
  display: block !important;
}

/* https://github.com/readthedocs/sphinx_rtd_theme/issues/1529 */
/* Take out pointless vertical whitespace in the signatures. */
.rst-content dl .sig dl,
.rst-content dl .sig dd {
  margin-bottom: 0;
}

/* Make signature boxes full-width, with view-source and header links right-aligned. */
.rst-content dl .sig {
  width: -webkit-fill-available;
}
.rst-content .viewcode-link {
  display: inline-flex;
  float: inline-end;
  margin-right: 1.5em;
}
.rst-content .headerlink {
  position: absolute;
  right: 0.5em;
}
